﻿/* 
    PASS Community Summit 2021
    Hacking SQL Server Another Day
    2021-11-10 (version 1.61)
    Demo - MDF file tampering
    André Melancia
    https://Andy.PT

    (C) 2016-2021, André Melancia
    
           ________   ________    _________  ____________;_
          - ______ \ - ______ \ / _____   //.  .  ._______/
         / /     / // /     / //_/     / // ___   /
        / /     / // /     / /       .-'//_/|_/,-'
       / /     / // /     / /     .-'.-'
      / /     / // /     / /     / /
     / /     / // /     / /     / /
    / /_____/ // /_____/ /     / /
    \________- \________-     /_/
         
*/

--  +----------------------------------------------------------------+
--  | MISSION BRIEFING                                               |
--  +----------------------------------------------------------------+

    -- You are an NSA agent already infiltrated in MI6.
    -- Your mission, if you decide to accept it, is to DELETE some
    -- NSA secrets stored in MI6 databases.
    -- After your mission preparation, you must run the demos in the
    -- order shown.
    -- As always, if you don't run these scripts on an EMPTY database,
    -- the Secretary will disavow any knowledge of your actions. 
    -- This script will NOT self-destruct in five seconds.
    
    -- WARNING:
    -- DON'T RUN THIS ON A PRODUCTION MACHINE. USE TEST MACHINES ONLY!
    
    -- These are ethical hacking scripts. 
    -- Do not use them for evil (pranking is fine).


--  +----------------------------------------------------------------+
--  | MISSION PREPARATION                                            |
--  +----------------------------------------------------------------+

    -- Drop old database (if any)
    USE Master;

    IF (EXISTS (SELECT name FROM master.dbo.sysdatabases WHERE (name = 'MI6Unencrypted')))
    BEGIN
        ALTER DATABASE MI6Unencrypted SET SINGLE_USER;
        DROP DATABASE IF EXISTS MI6Unencrypted;
    END --IF
    GO


    -- Create new database  
    CREATE DATABASE MI6Unencrypted;
    GO

    USE MI6Unencrypted;
    GO


    -- Create the Secrets table
    CREATE TABLE dbo.Secrets 
    (
        ID         INT IDENTITY(1,1) PRIMARY KEY,
        SecretText VARCHAR (255)
    )
    GO


--  +----------------------------------------------------------------+
--  | NORMAL OPERATION                                               |
--  +----------------------------------------------------------------+

    -- Insert A BIG secret!
    INSERT INTO dbo.Secrets (SecretText)
    VALUES ('AUSTIN POWERS is an NSA secret agent!');
    GO

    -- Insert a lot of secrets!
    INSERT INTO dbo.Secrets (SecretText)
    VALUES ('MAXWELL SMART is a CONTROL secret agent!');
    GO 100


    -- Confirm all the secrets
    SELECT   SecretText, Min(ID) AS ID_Min, Max(ID) AS ID_Max, Count(*) AS RecCount
    FROM     dbo.Secrets
    GROUP BY SecretText;
    GO

	
--  +----------------------------------------------------------------+
--  | DETACH DATABASE                                                |
--  +----------------------------------------------------------------+

    -- For whatever secret reason, M decided to detach this database.
    USE Master;
    GO
    ALTER DATABASE MI6Unencrypted
    SET SINGLE_USER;
    GO
    EXEC sp_detach_db 'MI6Unencrypted', 'true';
    GO


--  +----------------------------------------------------------------+
--  | IT'S NSA TAMPERING TIME                                        |
--  +----------------------------------------------------------------+

    -- Now the NSA agent will find the database,
    -- and make some changes to some bits and bytes.


--  +----------------------------------------------------------------+
--  | REATACH DATABASE                                               |
--  +----------------------------------------------------------------+

    -- M orders the database to be attached again
    EXEC sp_attach_db @dbname = N'MI6Unencrypted', 
                      @filename1 = N'C:\Program Files\Microsoft SQL Server\MSSQL15.MSSQLSERVER\MSSQL\DATA\MI6Unencrypted.MDF';
    GO

    USE MI6Unencrypted;
    GO


--  +----------------------------------------------------------------+
--  | READ THE INFORMATION                                           |
--  +----------------------------------------------------------------+

    -- Now M wants you to do something to the information...
    SELECT   SecretText, Min(ID) AS ID_Min, Max(ID) AS ID_Max, Count(*) AS RecCount
    FROM     dbo.Secrets
    GROUP BY SecretText;
    GO
    
    -- ... but SOMETHING'S WRONG.


--  +----------------------------------------------------------------+
--  | DBCC CHECKDB                                                   |
--  +----------------------------------------------------------------+

    -- Any DBA with licence to CHECKDB knows what to do...
    ALTER DATABASE MI6Unencrypted SET SINGLE_USER;
    GO
    DBCC CHECKDB ('MI6Unencrypted') /* WITH ALL_ERRORMSGS, DATA_PURITY */ ; 
    GO

    -- ERROR FOUND
    -- « CHECKDB found 0 allocation errors and NNN consistency errors in database 'MI6Unencrypted'.
    --   repair_allow_data_loss is the minimum repair level for the errors found by DBCC CHECKDB (MI6Unencrypted). »

    -- At this point, you just want it up and running as soon as possible:
    ALTER DATABASE MI6Unencrypted SET SINGLE_USER;
    GO
    DBCC CHECKDB ('MI6Unencrypted', REPAIR_ALLOW_DATA_LOSS)  WITH ALL_ERRORMSGS, DATA_PURITY  ;
    GO

    -- And THAT was your mistake!

--  +----------------------------------------------------------------+
--  | READING THE INFORMATION AGAIN                                  |
--  +----------------------------------------------------------------+

	ALTER DATABASE MI6Unencrypted
	SET MULTI_USER;
	GO

    USE MI6Unencrypted
    GO

    SELECT   SecretText, Min(ID) AS ID_Min, Max(ID) AS ID_Max, Count(*) AS RecCount
    FROM     dbo.Secrets
    GROUP BY SecretText;
    GO

    -- Some of the damaged records have been removed.
    -- No MI6 Secrets for you! NEXT!

--  +----------------------------------------------------------------+
--  | EXPERT MODE                                                    |
--  +----------------------------------------------------------------+

    -- You can also CHANGE the contents, but you need to make sure the
    -- checksums are correct. Another time, Mr.Bond...


--  +----------------------------------------------------------------+
--  | FIX IT                                                         |
--  +----------------------------------------------------------------+

    -- Use Transparent Data Encryption (TDE)
	-- https://docs.microsoft.com/sql/relational-databases/security/encryption/transparent-data-encryption

	-- USE master;
	-- GO
	-- CREATE MASTER KEY ENCRYPTION BY PASSWORD = '<UseStrongPasswordHere>';
	-- go
	-- CREATE CERTIFICATE MyServerCert WITH SUBJECT = 'My DEK Certificate';
	-- go
	-- USE AdventureWorks2012;
	-- GO
	-- CREATE DATABASE ENCRYPTION KEY
	-- WITH ALGORITHM = AES_256
	-- ENCRYPTION BY SERVER CERTIFICATE MyServerCert;
	-- GO
	-- ALTER DATABASE AdventureWorks2012
	-- SET ENCRYPTION ON;
	-- GO

    -- Setting TDE on a database will also encrypt TEMPDB!


--  +----------------------------------------------------------------+
--  | Mission cleanup                                                |
--  +----------------------------------------------------------------+

    -- You must delete any databases or junk you created on the 
    -- test server.


--  +----------------------------------------------------------------+
--  | Mission is now over.                                           |
--  | You can't run away with the Bond Girl because you're an NSA    |
--  | agent, not an MI6 agent.                                       |
--  | Consider changing sides: MI6 has better... eh... cars...       |
--  +----------------------------------------------------------------+
